/*
 * matilda.c:
 * Run a program, and pass on its output if it's failed on every occasion it's
 * been run for more than some set interval.
 *
 * Copyright (c) 2004 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: matilda.c,v 1.2 2004/06/13 12:32:22 chris Exp $";

#include <sys/types.h>

#include <ctype.h>
#include <errno.h>
#include <gdbm.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <openssl/md5.h>

#include <sys/wait.h>

/* parse_interval STRING
 * Interpret STRING, which matches /^\s*\d+\s*[mhdw]/, as an interval. Returns
 * the number of seconds in the interval on success, or 0 on failure. */
time_t parse_interval(const char *s) {
    const char *p;
    time_t a;
    p = s + strspn(s, " \t");
    if (!isdigit(*p))
        return 0;
    a = (time_t)atoi(p);
    p += strspn(p, "0123456789");
    p += strspn(p, " \t");
    if (*p) {
        switch (tolower(*p)) {
            case 'w':
                a *= 7 * 24 * 3600;
                break;

            case 'd':
                a *= 24 * 3600;
                break;

            case 'h':
                a *= 3600;
                break;

            case 'm':
                a *= 60;
                break;

            case 's':
                break;

            default:
                a = 0;
                break;
        }
    } else
        a *= 24 * 3600; /* assume `days' */

    return a;
}

/* run_program RESULT ARGV
 * Run the program in ARGV, saving the result (as returned from wait(2)) in
 * *RESULT, and return a stream opened on a file containing the output of the
 * program, or NULL on error. */
FILE *run_program(int *res, char *argv[]) {
    FILE *fp;
    pid_t p;

    if (!(fp = tmpfile()))
        return NULL;

    switch ((p = fork())) {
        int real_stderr, ee;
        
        case -1:
            fclose(fp);
            return NULL;

        case 0:
            /* Close standard output/standard error, and reopen them on the
             * temporary file. Then run the given program. This is complicated
             * by one feature, which is that if exec fails we want to report
             * this directly on the parent's standard error. This is a pain,
             * since it means that real standard error is still open in the
             * child, under some random file descriptor, but that's not too big
             * a problem. */
            real_stderr = dup(2);
            close(1);
            close(2);
            dup(fileno(fp));
            dup(fileno(fp));
            close(fileno(fp));
            
            errno = 0;
            if (-1 == execvp(argv[0], argv)) {
                FILE *e;
                ee = errno;
                if ((e = fdopen(real_stderr, "w"))) {
                    fprintf(e, "matilda: %s: execvp: %s\n", argv[0], strerror(ee));
                    fclose(e);
                }
            }
            
            _exit(1);

        default:
            wait(res);
            return fp;
    }
}

/* record_result STATUS ARGV
 * Record that running the program in ARGV completed with given STATUS (nonzero
 * is success); returns the last time at which the status changed, or -1 on
 * failure. The special return value 0 means that no previous status is
 * recorded. */
struct runstatus {
    time_t rs_time;
    int rs_status;
} record_result(int status, char *argv[]) {
    char *dbfile = NULL, *cwd = NULL, buf[32], *p, **a;
    unsigned char hash[16];
    size_t buflen;
    GDBM_FILE db = NULL;
    time_t lasttime = -1;
    int laststatus;
    MD5_CTX md5;
    datum key, val;
    struct runstatus R;

    /* 
     * Compute the MD5 checksum of the command and some other current state as
     * a key to record this command.
     */
    MD5_Init(&md5);

    /* Current directory. */
    cwd = malloc(buflen = 1024);
    while (!(p = getcwd(cwd, buflen)) && errno == ERANGE)
        cwd = realloc(cwd, buflen *= 2);
    if (!p)
        goto fail;

    MD5_Update(&md5, cwd, strlen(cwd) + 1);

    /* UID, GID. */
    sprintf(buf, "%lld", (long long int)getuid());
    MD5_Update(&md5, buf, strlen(buf) + 1);

    sprintf(buf, "%lld", (long long int)getgid());
    MD5_Update(&md5, buf, strlen(buf) + 1);
    
    /* Command-line. */
    for (a = argv; *a; ++a) {
        MD5_Update(&md5, *a, strlen(*a) + 1);
    }

    MD5_Final(hash, &md5);

    key.dptr = hash;
    key.dsize = 16;

    /*
     * Open the database.
     */
    if (!(dbfile = getenv("MATILDA_DB"))) {
        char *home;
        if (!(home = getenv("HOME"))) {
            struct passwd *p;
            if (!(p = getpwuid(getuid())))
                goto fail;
            home = p->pw_dir;
        }
        dbfile = malloc(strlen(home) + 1 + sizeof ".matildadb");
        sprintf(dbfile, "%s/.matildadb", home);
    } else
        dbfile = strdup(dbfile);

    gdbm_errno = 0;
    errno = 0;
    if (!(db = gdbm_open(dbfile, 0, GDBM_WRCREAT, 0644, NULL))) {
        fprintf(stderr, "matilda: %s: %s: gdbm_open: %s\n", dbfile, argv[0], gdbm_errno ? gdbm_strerror(gdbm_errno) : (errno ? strerror(errno) : "unknown error"));
        goto fail;
    }

    /* 
     * Obtain current state. This is recorded in the database in the form
     * /^[YN]\d+$/ where Y indicates that the last result was success, N that
     * it was failure, and the digit string records the time at which that
     * result was recorded.
     */
#define NOLASTSTATUS()  do { lasttime = 0; laststatus = !status; } while (0)
    val = gdbm_fetch(db, key);
    if (!val.dptr || val.dsize < 2)
        NOLASTSTATUS();
    else {
        char *s;
        s = (char*)val.dptr;
        if (strchr("YN", *s)) {
            int i;
            if (*s == 'Y')
                laststatus = 1;
            else
                laststatus = 0;
            for (i = 1, lasttime = 0; i < val.dsize; ++i) {
                if (!strchr("0123456789", s[i])) {
                    NOLASTSTATUS();
                    break;
                }
                lasttime = (lasttime * 10) + s[i] - '0';
            }
        } else
            NOLASTSTATUS();
        free(val.dptr);
    }
    
    /* 
     * Save new state if it is different from the last state.
     */
    if (laststatus != status) {
        sprintf(buf, "%c%lld", status ? 'Y' : 'N', (long long int)time(NULL));
        val.dptr = buf;
        val.dsize = strlen(buf);
        gdbm_store(db, key, val, GDBM_REPLACE);
    }
    
fail:

    if (cwd) free(cwd);
    if (dbfile) free(dbfile);
    if (db) gdbm_close(db);

    R.rs_time = lasttime;
    R.rs_status = laststatus;
    return R;
}

int main(int argc, char *argv[]) {
    time_t howlong;
    FILE *output;
    int res, status = 0;
    struct runstatus R;
   
    if (argv[1] && (0 == strcmp(argv[1], "-h") || 0 == strcmp(argv[1], "--help"))) {
        printf(
"matilda - run a program, but don't report its output unless every invocation\n"
"          of the program over a specified interval has failed.\n"
"\n"
"Usage: matilda INTERVAL PROGRAM ARG ...\n"
"\n"
"INTERVAL is of the form NUMBER [UNIT]. UNIT may be `seconds', `minutes',\n"
"`hours', `days', `weeks', or any unambiguous abbreviation. If not specified\n"
"`days' is assumed.\n"
"\n"
"Copyright (c) 2004 Chris Lightfoot <chris@ex-parrot.com>\n"
"http://ex-parrot.com/~chris/\n");
        return 0;
    }
    
    if (argc < 3) {
        fprintf(stderr, "matilda: must have at least two arguments\n");
        return 1;
    }
    
    /* Arguments are interval, and program to run. */
    if (!(howlong = parse_interval(argv[1]))) {
        fprintf(stderr, "matilda: %s: %s: not a valid interval\n", argv[2], argv[1]);
        return 1;
    }

    if (!(output = run_program(&res, argv + 2))) {
        fprintf(stderr, "matilda: %s: unable to run: %s\n", argv[2], strerror(errno));
        return 1;
    }

    if (WIFEXITED(res) && 0 == WEXITSTATUS(res))
        status = 1;

    R = record_result(status, argv + 2);
    if (!status && R.rs_status) {
        int do_output = 1;
        if (-1 == R.rs_time)
            fprintf(stderr, "matilda: %s: unable to access last-status database; output of program follows:\n", argv[2]);
        else if (0 == R.rs_time)
            fprintf(stderr, "matilda: %s: failure first time program run; output of program follows:\n", argv[2]);
        else if (R.rs_time < time(NULL) - howlong) {
            struct tm *tm;
            char tbuf[32];
            tm = localtime(&R.rs_time);
            strftime(tbuf, sizeof tbuf, "%Y-%m-%dT%H:%M:%S", tm);
            fprintf(stderr, "matilda: %s: program hasn't completed successfully since %s; last output follows:\n", argv[2], tbuf);
        } else
            do_output = 0;

        if (do_output) {
            int c;
            rewind(output);
            while (EOF != (c = fgetc(output))) {
                if (EOF == fputc(c, stdout)) {
                    fprintf(stderr, "matilda: %s: fputc(output): %s\n", argv[2], strerror(errno));
                    return 1;
                }
                if (ferror(output)) {
                    fprintf(stderr, "matilda: %s: fgetc(tempfile): %s\n", argv[2], strerror(errno));
                    return 1;
                }
            }
        }
    } else
        return 0;

    if (WIFEXITED(res))
        return WEXITSTATUS(res);
    else {
        /* Yuck: */
        signal(WTERMSIG(res), SIG_DFL);
        raise(WTERMSIG(res));
        /* Hopefully this program will now be killed with whatever signal killed
         * the program we ran. If not, oh well... */
        return 1;
    }
}
